#include <stdio.h>
#include "typedefs.h"

/*** <white space optimization> ***/

/* function the compact sequence white spaces in a string to one space
it changes the character pointer
the function returns the length of the compacted string*/
int CompactWhiteSpacesInString(char **str)
{
    int len = strlen(*str);
    char *inStr = *str;
    char *newStr = (char *)safeMalloc((len + 1) * sizeof(char));
    char *ret = newStr;

    len = 0;
    while(*inStr != '\0')
    {
        if(isspace(*inStr))
        {
            /* put one space */
            *newStr++ = ' ';
            len++;
            while(isspace(*inStr))
                inStr++;
        }
        else
        {
            *newStr = *inStr;
            newStr++; inStr++;
            len++;
        }
    }
    *newStr = '\0';
    free(*str);
    *str = ret;
    return len;
}

/* function to do one step of whitespacecompression andsimplification.
Writing thisas a separate step because we might want to combine it with
some other pass */
void WhiteSpaceStep(NODE *node)
{
    int len;

    if(node == NULL)
        return;

    /* do nothing if thisisn't a text node, or is in TT */
    if(node->type != TEXT || node->in.tt)
        return;

    len = CompactWhiteSpacesInString(&node->s);

    /* if last char in the string is a space, and the first in the next text
    is also a text, we can remove the ending space from this string */
    if(isspace(node->s[len-1]))
    {
        if(node->nexttext != NULL)
        {
            if(isspace(node->nexttext->s[0]))
            {
                /* do it only if relevant parts of context are same */
                if(isequal(SpaceContext(node->in), SpaceContext(node->nexttext->in)))
                {
                    node->s[len-1] = '\0';
                }
            }
        }
    }
}

/* function to perform whitespace compression and simplification on node list */
void WhiteSpaceOptimization(NODE *head)
{
    while(head)
    {
        WhiteSpaceStep(head);
        head = head->next;
    }
}

/*** </white space optimization> ***/

/*** <equivalence check> ***/

/* function that gives the next character of text in corresponding SML/NG.
t takes a pointer to a NODE *(it might have to changes the node),and a pointer to an integer.
This integer is the index of the "current character"in node->s.
The boolean wasNodeChanged, tells us whether, this next character,
belongs to a different node*/
void GetNextChar(NODE ** ppNode,int *i, int *wasNodeChanged)
{
    NODE *pNode = *ppNode;
    *wasNodeChanged = 0;

    if(pNode == NULL)
        return;

    if(pNode->type != TEXT)
        myError("GetNextChar called with non text node");

    /* see if the next character is in the next node */
    if(pNode->s[*i+1] == '\0')
    {
        *i = 0;
        *ppNode = pNode->nexttext;
        *wasNodeChanged = 1;
        return;
    }
    else
    {
        (*i)++ ;
        *wasNodeChanged = 0;
    }
}

/* this function is similar to GetNextChar above, but also considers a sequence of whitespaces
as one whitespace, if the context is not TT */
void GetNextCharW(NODE ** ppNode,int *i, int *wasNodeChanged)
{
    NODE *pNode = *ppNode;

    if(isspace(pNode->s[*i]))
    {
        /* if context is not TT skip all subsequent consecutive whitespace characters */
        GetNextChar(ppNode, i, wasNodeChanged);

        while(*ppNode && !(*ppNode)->in.tt && isspace((*ppNode)->s[*i]))
            GetNextChar(ppNode, i, wasNodeChanged);

    }
    else
    {
        GetNextChar(ppNode, i, wasNodeChanged);
    }
}

/*** <copying node lists> ***/
/* function to copy a node list */
NODE *CopyNodeList(NODE *head)
{
    NODE *newHead = NULL;
    NODE *prevn = NULL, *prevtextn = NULL;
    NODE *node = NULL;
    NODESTK *nodeStk = NULL;

    while(head)
    {
        node = newNode();
        if(newHead == NULL)
            newHead = node;

        *node = *head;
        node->prev = prevn;
        if(prevn != NULL)
            prevn->next = node;

        prevn = node;

        if(node->type == TEXT)
        {
            node->prevtext = prevtextn;
            if(prevtextn != NULL)
                prevtextn->nexttext = node;
            prevtextn = node;

            node->s = safeMalloc(sizeof(char) * (strlen(head->s)+1));
            strcpy(node->s, head->s);
        }
        else if(isstarttag(node->type))
            nodeStk = push(nodeStk, node);
        else if(isendtag(node->type))
        {
            NODE *matchNode = pop(&nodeStk);
            matchNode->match = node;
            node->match = matchNode;
        }

        head = head->next;
    }
    return newHead;
}
/*** </copying node lists> ***/

/*** <size of file corresponding to a node list ***/
/* function that returns the size of file corresponding to nodeList */
int NodeSize(NODE *node)
{
    char *outbuf = NULL;
    switch(node->type)
    {
    case TEXT:
        outbuf = node->s;
        break;
    case BOLD:
        outbuf = "<B>";
        break;
    case BOLD_END:
        outbuf = "</B>";
        break;
    case EMPHASIS:
        outbuf = "<EM>";
        break;
    case EMPHASIS_END:
        outbuf = "</EM>";
        break;
    case ITALICS:
        outbuf = "<I>";
        break;
    case ITALICS_END:
        outbuf = "</I>";
        break;
    case PLAIN:
        outbuf = "<PL>";
        break;
    case PLAIN_END:
        outbuf = "</PL>";
        break;
    case STRONG:
        outbuf = "<S>";
        break;
    case STRONG_END:
        outbuf = "</S>";
        break;
    case TT:
        outbuf = "<TT>";
        break;
    case TT_END:
        outbuf = "</TT>";
        break;
    case UNDERLINE:
        outbuf = "<U>";
        break;
    case UNDERLINE_END:
        outbuf = "</U>";
        break;
    case SIZE_0:
        outbuf = "<0>";
        break;
    case SIZE_0_END:
        outbuf = "</0>";
        break;
    case SIZE_1:
        outbuf = "<1>";
        break;
    case SIZE_1_END:
        outbuf = "</1>";
        break;
    case SIZE_2:
        outbuf = "<2>";
        break;
    case SIZE_2_END:
        outbuf = "</2>";
        break;
    case SIZE_3:
        outbuf = "<3>";
        break;
    case SIZE_3_END:
        outbuf = "</3>";
        break;
    case SIZE_4:
        outbuf = "<4>";
        break;
    case SIZE_4_END:
        outbuf = "</4>";
        break;
    case SIZE_5:
        outbuf = "<5>";
        break;
    case SIZE_5_END:
        outbuf = "</5>";
        break;
    case SIZE_6:
        outbuf = "<6>";
        break;
    case SIZE_6_END:
        outbuf = "</6>";
        break;
    case SIZE_7:
        outbuf = "<7>";
        break;
    case SIZE_7_END:
        outbuf = "</7>";
        break;
    case SIZE_8:
        outbuf = "<8>";
        break;
    case SIZE_8_END:
        outbuf = "</8>";
        break;
    case SIZE_9:
        outbuf = "<9>";
        break;
    case SIZE_9_END:
        outbuf = "</9>";
        break;

    case RED:
        outbuf = "<r>";
        break;
    case RED_END:
        outbuf = "</r>";
        break;
    case GREEN:
        outbuf = "<g>";
        break;
    case GREEN_END:
        outbuf = "</g>";
        break;
    case BLUE:
        outbuf = "<b>";
        break;
    case BLUE_END:
        outbuf = "</b>";
        break;
    case CYAN:
        outbuf = "<c>";
        break;
    case CYAN_END:
        outbuf = "</c>";
        break;
    case MAGENTA:
        outbuf = "<m>";
        break;
    case MAGENTA_END:
        outbuf = "</m>";
        break;
    case YELLOW:
        outbuf = "<y>";
        break;
    case YELLOW_END:
        outbuf = "</y>";
        break;
    case BLACK:
        outbuf = "<k>";
        break;
    case BLACK_END:
        outbuf = "</k>";
        break;
    case WHITE:
        outbuf = "<w>";
        break;
    case WHITE_END:
        outbuf = "</w>";
        break;
    default:
        myError("Bad node type in NodeSize");
    }
    return strlen(outbuf);
}

/* function that gives the size of the output file corresponding to a node list */
int NodeListSize(NODE *head)
{
    int size = 0;
    while(head != NULL)
    {
        size += NodeSize(head);
        head = head->next;
    }
    return size;
}
/*** </size of file corresponding to a node list ***/


/* Given two node lists, this function tells whether the meaning of the text corresponding
to them is the same.
Returns 1 if equivalent, 0 otherwise
*/
int IsEquivalent(NODE *head1, NODE *head2)
{
	int i1, i2;

	i1 = 0;
	i2 = 0;

	/* first time, skip till we come to a text node */
	while(head1 != NULL && head1->type != TEXT)
		head1 = head1->next;
	while(head2 != NULL && head2->type != TEXT)
		head2 = head2->next;

	/* if only one of them is NULL return 0 */
	if(head1 == NULL && head2 == NULL)
		return 1;
	if(head1 == NULL || head2 == NULL)
		return 0;

	if(!isequal(head1->in, head2->in) || head1->s[i1] != head2->s[i2])
		return 0;

	while(1)
	{
		int wasHead1Changed, wasHead2Changed;

		GetNextCharW(&head1, &i1, &wasHead1Changed);
		GetNextCharW(&head2, &i2, &wasHead2Changed);

		/* if only one of them is NULL return 0 */
		if(head1 == NULL && head2 == NULL)
			return 1;

		if(head1 == NULL || head2 == NULL)
			return 0;

		/* if text char are diff, return 0 */
		if(head1->s[i1] != head2->s[i2])
			return 0;

		if(wasHead1Changed || wasHead2Changed)
		{
			/* check the contexts exactly only if they are non space characters.
			This is because we skip white spaces and the node may change in between and
			context won't matter */
			if(!isspace(head1->s[i1]))
			{
				if(!isequal(head1->in, head2->in))
					return 0;
			}
			else if(!isequal(SpaceContext(head1->in), SpaceContext(head2->in)))				
				return 0;

		}
	}
	return 1;
}
/*** </equivalence check> ***/

/*** <dependency analysis> ***/
/* Basically we'll use a mark sweep garbage collection scheme. 
In the mark pass, all nodes pointed to by a text nodes context will be marked useful.
In the sweep pass all nodes not marked useful will be removed.
Note that EM tags are NEVER to be removed in this
*/

static void MarkUsefulNodes(CONTEXT ctxt)
{
	if(ctxt.n_b != NULL)
		ctxt.n_b->useful = 1;
	if(ctxt.n_i != NULL)
		ctxt.n_i->useful = 1;
	if(ctxt.n_s != NULL)
		ctxt.n_s->useful = 1;
	if(ctxt.n_tt != NULL)
		ctxt.n_tt->useful = 1;
	if(ctxt.n_u1 != NULL)
		ctxt.n_u1->useful = 1;
	if(ctxt.n_u2 != NULL)
		ctxt.n_u2->useful = 1;
	if(ctxt.n_u3 != NULL)
		ctxt.n_u3->useful = 1;
	if(ctxt.n_size != NULL)
		ctxt.n_size->useful = 1;
	if(ctxt.n_color != NULL)
		ctxt.n_color->useful = 1;
}

void Dep_Mark(NODE *head)
{
	NODE *p = head;

	/* skip till the first text node */
	while(p != NULL && p->type != TEXT)
		p = p->next;

	while(p != NULL)
	{
		MarkUsefulNodes(p->in);
		p = p->nexttext;
	}
}


NODE *Dep_Sweep(NODE *head)
{
	NODE *p = head;

	while(p)
	{
		if(isstarttag(p->type) && p->type != EMPHASIS && !p->useful)
		{
			NODE *saveNext;
			saveNext = p->next;
			if(saveNext == p->match)
				saveNext = saveNext->next;

			/* off with the useless node! */
			head = RemoveTagPair(head, p);
			p = saveNext;
		}
		else
		{
			p = p->next;
		}
	}
	return head;
}

NODE *Dep_GarbageCollect(NODE *head)
{
	Dep_Mark(head);
	return Dep_Sweep(head);
}
/*** </dependency analysis> ***/
