#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* text */

#define TEXT 1

/* beginning tags */

#define BOLD 2
#define EMPHASIS 3
#define ITALICS 4
#define PLAIN 5
#define STRONG 6
#define TT 7
#define UNDERLINE 8
#define SIZE_0 9
#define SIZE_1 10
#define SIZE_2 11
#define SIZE_3 12
#define SIZE_4 13
#define SIZE_5 14
#define SIZE_6 15
#define SIZE_7 16
#define SIZE_8 17
#define SIZE_9 18
#define RED 19
#define GREEN 20
#define BLUE 21
#define CYAN 22
#define MAGENTA 23
#define YELLOW 24
#define BLACK 25
#define WHITE 26

/* macro to test if a given type corresponds to a starting tag */
#define isstarttag(a) ((a) >= BOLD && (a) <= WHITE)

/* ending tags */

#define BOLD_END 27
#define EMPHASIS_END 28
#define ITALICS_END 29
#define PLAIN_END 30
#define STRONG_END 31
#define TT_END 32
#define UNDERLINE_END 33
#define SIZE_0_END 34
#define SIZE_1_END 35
#define SIZE_2_END 36
#define SIZE_3_END 37
#define SIZE_4_END 38
#define SIZE_5_END 39
#define SIZE_6_END 40
#define SIZE_7_END 41
#define SIZE_8_END 42
#define SIZE_9_END 43
#define RED_END 44
#define GREEN_END 45
#define BLUE_END 46
#define CYAN_END 47
#define MAGENTA_END 48
#define YELLOW_END 49
#define BLACK_END 50
#define WHITE_END 51

/* macro to test if a given type corresponds to an ending tag */
#define isendtag(a) ((a) >= BOLD_END && (a) <= WHITE_END)

struct node;
typedef struct node NODE;

/* in the context, we not only have fields that store the context info,
we also have (for every field except EM), a pointer to the start tag node who caused
that change in the context */
typedef struct
{
    int b, em, i, s, tt;         /* 0 false, 1 true */
	NODE *n_b, *n_i, *n_s, *n_tt;

    int u, size, color;
	NODE *n_u1, *n_u2, *n_u3, *n_size, *n_color;

} CONTEXT;                       /* see problem description */


struct node
{
    CONTEXT in;          /* context before this node (incoming context) */
    int type;            /* TEXT or any of the above 50 tags */

    struct node *prev;   /* pointer to the previous node */
    struct node *next;   /* pointer to the next node */

    /*
     * the following pointers matter only if type == TEXT
     * if type != TEXT, they should be set to NULL
     */

    char *s;                  /* text string */
    struct node *prevtext;    /* prev text (ignore tags in between) */
    struct node *nexttext;    /* next text (ignore tags in between) */

    /*
     * the following pointers matter only if type != TEXT (i.e., a tag)
     * if type == TEXT, they should be set to NULL
     */

    struct node *match;       /* the closing/starting matching tag */

	int useful;	/* this is a flag used in dependency detection garbage collection */
};


/* node stack primitives */
typedef
struct tNodeStk
{
    struct tNodeStk *down;
    NODE *item;
} NODESTK;

NODESTK *push(NODESTK *oldStk, NODE *item);
NODE *pop(NODESTK **stk);

/* function declarations */
NODE *newNode();
NODE *List_remove(NODE *, NODE *);
NODE *EM_elim(NODE *);
NODE *rem_redundant(NODE *);
NODE *PL_ShortCut( NODE *);
NODE *RemoveTagPair(NODE *head, NODE *node);
CONTEXT AddTagToContext(CONTEXT *in, NODE *node);
int isequal(CONTEXT c1, CONTEXT c2);
void WhiteSpaceStep(NODE *node);
void WhiteSpaceOptimization(NODE *head);
NODE *CopyNodeList(NODE *head);
int NodeListSize(NODE *head);
int IsEquivalent(NODE *head1, NODE *head2);
NODE *List_InsertAfterNode( NODE *InsertPoint, int Type, char *TextString );
void AdjustContext( NODE *StartNode );
NODE *PL_ShortCut( NODE *root );
NODE *Dep_GarbageCollect(NODE *head);
void *safeMalloc(size_t size);
void myError(char *msg);
CONTEXT SpaceContext(CONTEXT ctxt);

